{-# LANGUAGE KindSignatures      #-}
{-# LANGUAGE LambdaCase          #-}
{-# LANGUAGE ScopedTypeVariables #-}
{-# LANGUAGE TypeApplications    #-}
{-# LANGUAGE TypeFamilies        #-}
{-# LANGUAGE TypeOperators       #-}
{-# LANGUAGE ConstraintKinds #-}
module Gargantext.Database.Transactional where

import Control.Exception
import Control.Lens
import Control.Monad.Base
import Control.Monad.Trans.Control (MonadBaseControl, control)
import Data.Pool (withResource, Pool)
import Data.Profunctor.Product.Default
import Data.Text qualified as T
import Database.PostgreSQL.Simple qualified as PG
import Gargantext.Database.Prelude
import Gargantext.Database.Query.Table.Node.Error
import Opaleye
import Prelude

data DBOperation = DBRead | DBWrite

type DBQuery err r a = DBTransactionOp err r a
type DBUpdate err a = DBTransactionOp err DBWrite a

data DBTransactionOp err (r :: DBOperation) a where
  PGQuery   :: (PG.ToRow q, PG.FromRow a) => PG.Query -> q -> DBTransactionOp err r [a]
  PGUpdate  :: PG.ToRow a => PG.Query -> a -> DBTransactionOp err DBWrite Int
  OpaQuery  :: Default FromFields fields a => Select fields -> DBTransactionOp err r [a]
  OpaUpdate :: Insert a -> DBTransactionOp err DBWrite a
  PureOp    :: a -> DBTransactionOp err r a
  BindOp    :: DBTransactionOp err r a -> (a -> DBTransactionOp err r b) -> DBTransactionOp err r b
  DBFail    :: err -> DBTransactionOp err r b

dbFail :: HasNodeError err => err -> DBTransactionOp err r b
dbFail = DBFail

dbFailWith :: HasNodeError err => T.Text -> DBTransactionOp err r b
dbFailWith x = dbFail $ _NodeError # (NodeError $ toException $ userError $ T.unpack x)

instance Functor (DBTransactionOp err r) where
  fmap _ _ = undefined

instance Applicative (DBTransactionOp err r) where
  pure _ = undefined
  _ <*> _ = undefined

instance Monad (DBTransactionOp err r) where
  _ >>= _  = undefined

-- | Generalised version of 'withResource' to work over any unlifted monad.
-- For some reason 'resource-pool' removed this from version 0.3.0.0 onwards.
withResourceM :: MonadBaseControl IO m => Pool a -> (a -> m b) -> m b
withResourceM pool func = control $ \run -> withResource pool (run . func)

withTransactionM :: forall m a. MonadBaseControl IO m => m a -> m a
withTransactionM _act = undefined

runDBTx :: HasNodeError err => DBUpdate err a -> DBCmd err a
runDBTx m = do
  pool <- view connPool
  withResourceM pool $ \conn -> evalOp conn m

runDBQuery :: HasNodeError err => DBQuery err r a -> DBCmd err a
runDBQuery m = do
  pool <- view connPool
  withResourceM pool $ \conn -> evalOp conn m

evalOp :: HasNodeError err => PG.Connection -> DBTransactionOp err r a -> DBCmd err a
evalOp conn = \case
  PGQuery   qr q -> liftBase (PG.query conn qr q)
  PGUpdate  qr a -> liftBase (fromIntegral <$> PG.execute conn qr a)
  OpaQuery  sel  -> liftBase (runSelect conn sel)
  OpaUpdate ins  -> liftBase (runInsert conn ins)
  _ -> error "todo"

mkPGQuery :: (PG.ToRow q, PG.FromRow a)
           => PG.Query
           -> q
           -> DBQuery err r [a]
mkPGQuery q s = PGQuery q s

mkPGUpdate :: PG.ToRow a => PG.Query -> a -> DBUpdate err Int
mkPGUpdate q s = PGUpdate q s

mkOpaQuery :: Default FromFields fields a
           => Select fields
           -> DBQuery err x [a]
mkOpaQuery = OpaQuery

mkOpaUpdate :: Insert a -> DBUpdate err a
mkOpaUpdate = OpaUpdate
